@@ -4,6 +4,8 @@
 #pragma GCC diagnostic ignored "-Wunknown-warning-option"
 #pragma GCC diagnostic ignored "-Wuninitialized-const-reference"
 #pragma GCC diagnostic ignored "-Wdeprecated-declarations" // TODO: https://github.com/BinomialLLC/basis_universal/pull/333
+#pragma GCC diagnostic ignored "-Wbitwise-instead-of-logical"
+#pragma GCC diagnostic ignored "-Wdeprecated-builtins"
 #endif
 
 #ifdef __GNUC__
