@@ -303,12 +303,15 @@ static void encodeExpShared(uint32_t v[3], const Attr& a, int bits)
 	v[2] = (mz & mmask) | (unsigned(exp) << 24);
 }
 
-static uint32_t encodeExpOne(float v, int bits)
+static uint32_t encodeExpOne(float v, int bits, int min_exp = -100)
 {
 	// extract exponent
 	int e;
 	frexp(v, &e);
 
+	// clamp exponent to ensure it's valid after bias
+	e = std::max(e, min_exp);
+
 	// scale the mantissa to make it a K-bit signed integer (K-1 bits for magnitude)
 	int exp = e - (bits - 1);
 
@@ -321,49 +324,47 @@ static uint32_t encodeExpOne(float v, int bits)
 	return (m & mmask) | (unsigned(exp) << 24);
 }
 
-static void encodeExpParallel(std::string& bin, const Attr* data, size_t count, int bits)
+static void encodeExpParallel(std::string& bin, const Attr* data, size_t count, int channels, int bits, int min_exp = -100)
 {
-	int expx = -128, expy = -128, expz = -128;
+	int exp[4] = {};
+
+	for (int k = 0; k < channels; ++k)
+		exp[k] = min_exp;
 
 	for (size_t i = 0; i < count; ++i)
 	{
 		const Attr& a = data[i];
 
-		// get exponents from all components
-		int ex, ey, ez;
-		frexp(a.f[0], &ex);
-		frexp(a.f[1], &ey);
-		frexp(a.f[2], &ez);
-
 		// use maximum exponent to encode values; this guarantees that mantissa is [-1, 1]
-		expx = std::max(expx, ex);
-		expy = std::max(expy, ey);
-		expz = std::max(expz, ez);
+		for (int k = 0; k < channels; ++k)
+		{
+			int e;
+			frexp(a.f[k], &e);
+			exp[k] = std::max(exp[k], e);
+		}
 	}
 
 	// scale the mantissa to make it a K-bit signed integer (K-1 bits for magnitude)
-	expx -= (bits - 1);
-	expy -= (bits - 1);
-	expz -= (bits - 1);
+	for (int k = 0; k < channels; ++k)
+		exp[k] -= (bits - 1);
 
 	for (size_t i = 0; i < count; ++i)
 	{
 		const Attr& a = data[i];
 
-		// compute renormalized rounded mantissas
-		int mx = int(ldexp(a.f[0], -expx) + (a.f[0] >= 0 ? 0.5f : -0.5f));
-		int my = int(ldexp(a.f[1], -expy) + (a.f[1] >= 0 ? 0.5f : -0.5f));
-		int mz = int(ldexp(a.f[2], -expz) + (a.f[2] >= 0 ? 0.5f : -0.5f));
+		uint32_t v[4];
 
-		int mmask = (1 << 24) - 1;
+		for (int k = 0; k < channels; ++k)
+		{
+			// compute renormalized rounded mantissas
+			int m = int(ldexp(a.f[k], -exp[k]) + (a.f[k] >= 0 ? 0.5f : -0.5f));
 
-		// encode exponent & mantissa
-		uint32_t v[3];
-		v[0] = (mx & mmask) | (unsigned(expx) << 24);
-		v[1] = (my & mmask) | (unsigned(expy) << 24);
-		v[2] = (mz & mmask) | (unsigned(expz) << 24);
+			// encode exponent & mantissa
+			int mmask = (1 << 24) - 1;
+			v[k] = (m & mmask) | (unsigned(exp[k]) << 24);
+		}
 
-		bin.append(reinterpret_cast<const char*>(v), sizeof(v));
+		bin.append(reinterpret_cast<const char*>(v), sizeof(uint32_t) * channels);
 	}
 }
 
@@ -405,7 +406,7 @@ StreamFormat writeVertexStream(std::string& bin, const Stream& stream, const Qua
 
 			if (settings.compressmore)
 			{
-				encodeExpParallel(bin, &stream.data[0], stream.data.size(), qp.bits + 1);
+				encodeExpParallel(bin, &stream.data[0], stream.data.size(), 3, qp.bits + 1);
 			}
 			else
 			{
@@ -511,24 +512,66 @@ StreamFormat writeVertexStream(std::string& bin, const Stream& stream, const Qua
 		if (!settings.quantize)
 			return writeVertexStreamRaw(bin, stream, cgltf_type_vec2, 2);
 
-		float uv_rscale[2] = {
-		    qt.scale[0] == 0.f ? 0.f : 1.f / qt.scale[0],
-		    qt.scale[1] == 0.f ? 0.f : 1.f / qt.scale[1],
-		};
-
-		for (size_t i = 0; i < stream.data.size(); ++i)
+		if (settings.tex_float)
 		{
-			const Attr& a = stream.data[i];
+			StreamFormat::Filter filter = settings.compress ? StreamFormat::Filter_Exp : StreamFormat::Filter_None;
 
-			uint16_t v[2] = {
-			    uint16_t(meshopt_quantizeUnorm((a.f[0] - qt.offset[0]) * uv_rscale[0], qt.bits)),
-			    uint16_t(meshopt_quantizeUnorm((a.f[1] - qt.offset[1]) * uv_rscale[1], qt.bits)),
-			};
-			bin.append(reinterpret_cast<const char*>(v), sizeof(v));
+			// expand the encoded range to ensure it covers [0..1) interval
+			// this can slightly reduce precision but we should not need more precision inside 0..1, and this significantly
+			// improves compressed size when using encodeExpOne
+			const int min_exp = 0;
+
+			if (settings.compressmore)
+			{
+				encodeExpParallel(bin, &stream.data[0], stream.data.size(), 2, qt.bits + 1, min_exp);
+			}
+			else
+			{
+				for (size_t i = 0; i < stream.data.size(); ++i)
+				{
+					const Attr& a = stream.data[i];
+
+					if (filter == StreamFormat::Filter_Exp)
+					{
+						uint32_t v[2];
+						v[0] = encodeExpOne(a.f[0], qt.bits + 1, min_exp);
+						v[1] = encodeExpOne(a.f[1], qt.bits + 1, min_exp);
+						bin.append(reinterpret_cast<const char*>(v), sizeof(v));
+					}
+					else
+					{
+						float v[2] = {
+						    meshopt_quantizeFloat(a.f[0], qt.bits),
+						    meshopt_quantizeFloat(a.f[1], qt.bits)};
+						bin.append(reinterpret_cast<const char*>(v), sizeof(v));
+					}
+				}
+			}
+
+			StreamFormat format = {cgltf_type_vec2, cgltf_component_type_r_32f, false, 8, filter};
+			return format;
 		}
+		else
+		{
+			float uv_rscale[2] = {
+			    qt.scale[0] == 0.f ? 0.f : 1.f / qt.scale[0],
+			    qt.scale[1] == 0.f ? 0.f : 1.f / qt.scale[1],
+			};
 
-		StreamFormat format = {cgltf_type_vec2, cgltf_component_type_r_16u, qt.normalized, 4};
-		return format;
+			for (size_t i = 0; i < stream.data.size(); ++i)
+			{
+				const Attr& a = stream.data[i];
+
+				uint16_t v[2] = {
+				    uint16_t(meshopt_quantizeUnorm((a.f[0] - qt.offset[0]) * uv_rscale[0], qt.bits)),
+				    uint16_t(meshopt_quantizeUnorm((a.f[1] - qt.offset[1]) * uv_rscale[1], qt.bits)),
+				};
+				bin.append(reinterpret_cast<const char*>(v), sizeof(v));
+			}
+
+			StreamFormat format = {cgltf_type_vec2, cgltf_component_type_r_16u, qt.normalized, 4};
+			return format;
+		}
 	}
 	else if (stream.type == cgltf_attribute_type_normal)
 	{
@@ -758,6 +801,34 @@ StreamFormat writeVertexStream(std::string& bin, const Stream& stream, const Qua
 			return format;
 		}
 	}
+	else if (stream.type == cgltf_attribute_type_custom)
+	{
+		// note: _custom is equivalent to _ID, as such the data contains scalar integers
+		if (!settings.compressmore)
+			return writeVertexStreamRaw(bin, stream, cgltf_type_scalar, 1);
+
+		unsigned int maxv = 0;
+
+		for (size_t i = 0; i < stream.data.size(); ++i)
+			maxv = std::max(maxv, unsigned(stream.data[i].f[0]));
+
+		// exp encoding uses a signed mantissa with only 23 significant bits; input glTF encoding may encode indices losslessly up to 2^24
+		if (maxv >= (1 << 23))
+			return writeVertexStreamRaw(bin, stream, cgltf_type_scalar, 1);
+
+		for (size_t i = 0; i < stream.data.size(); ++i)
+		{
+			const Attr& a = stream.data[i];
+
+			uint32_t id = uint32_t(a.f[0]);
+			uint32_t v = id; // exp encoding of integers in [0..2^23-1] range is equivalent to the integer itself
+
+			bin.append(reinterpret_cast<const char*>(&v), sizeof(v));
+		}
+
+		StreamFormat format = {cgltf_type_scalar, cgltf_component_type_r_32f, false, 4, StreamFormat::Filter_Exp};
+		return format;
+	}
 	else
 	{
 		return writeVertexStreamRaw(bin, stream, cgltf_type_vec4, 4);
